#pragma once

#include <stdint.h>
#include <applibs/uart.h>
#include "eventloop_timer_utilities.h"

#define STR_LEN 32
#define MAX_APN_LEN 64
#define MAX_ICCID_LEN 20
#define MAX_IMEI_LEN 15
#define MAX_CONNECTED_SAT_NUMBER 2
#define MAX_GPS_AT_RESPONSE 512

#define USE_NXP_QUECTEL

typedef enum {
    network_type_disconnect = 0,
    network_type_unknown = 1,
    network_type_ethernet,
    network_type_wifi,
    network_type_4G,
    network_type_butt
} e_network_type;

enum AccessTech {
    AUTOMATIC = 0,  // LTE/WCDMA/TD-SCDMA/GSM for EC21  
    GSM,
    WCDMA,          // For BG96 NB-M or WCDMA for EC21
    LTE,            // For BG96 or TD-SCDMA for EC21, including LTE_M and LTE_NB
    TDSCDMA,
    UMTS,
    CDMAOnly,
    HDROnly,
    CDMAHDROnly
};

typedef struct esim_id
{
    char IMI[MAX_IMEI_LEN+1];
    char QCCID[MAX_ICCID_LEN+1];
} ESIM_ID_S;

typedef struct gps_snr
{
    char ncnt[STR_LEN];
    char snrs[STR_LEN];
} GPS_SNR_INFO_S;

typedef struct gps
{
    char utc[STR_LEN];
    char longitude[STR_LEN];
    char latitude[STR_LEN];
    char altitude[STR_LEN];
    char nsat[MAX_CONNECTED_SAT_NUMBER+1];
} GPS_INFO_S;

typedef struct cellular_config
{
    char apn[MAX_APN_LEN+1];
    char uname[STR_LEN];
    char psw[STR_LEN];
    char pin[STR_LEN];
    char number[STR_LEN];//Reserved for later PPP dialing compatibility
	bool dhcpser;
    int accessTech;
} CELLULAR_CONFIG_S;

typedef struct cellular_configPara
{
    CELLULAR_CONFIG_S cellular_parameter;
    bool config_flag;
} CELLULAR_CONFIG_P;

typedef struct cellular_status
{
    char state[16];
    char signal_quality[8];
    char access_tech[12];
    char imei[MAX_IMEI_LEN + 1];
    char registration_state[16];
    char operator_code[8];
    char operator_name[20];
    char app_ver[STR_LEN];
} CELLULAR_STATUS_S;

typedef struct wwan_info
{
    char connected[STR_LEN];
    char suspended[STR_LEN];
    char wwan_if[STR_LEN];
    char ip_timeout[STR_LEN];
    char apn[STR_LEN];
    char roaming[STR_LEN];
    char method[STR_LEN];
    char address[STR_LEN];
    char prefix[STR_LEN];
    char gateway[STR_LEN];
    char dns[STR_LEN];
    char mtu[STR_LEN];
} WWAN_INFO_S;

typedef struct cellular_sys_info
{
    char platform[STR_LEN];
    char osver[STR_LEN];
    char appver[STR_LEN];
    char buildtime[STR_LEN];
    char bootmedia[STR_LEN];
} CELLULAR_SYS_INFO_S;

typedef enum {
    LogLevel_Err = 0,
    LogLevel_Warn,
    LogLevel_Info,
    LogLevel_Debug,

    LogLevel_BUTT
} LogLevel_E;

void log_impl(int level, const char* file, const char* func,
    const int line, const char* format, ...);
#define LOG_IMPL(level, format, ...) log_impl(level, __FILE__, __FUNCTION__, __LINE__, format, ##__VA_ARGS__)
/* LOGX macros implemented by Log_Debug */
#define LOGI(format, ...) LOG_IMPL(LogLevel_Info, format, ##__VA_ARGS__)
#define LOGW(format, ...) LOG_IMPL(LogLevel_Warn, format, ##__VA_ARGS__)
#define LOGE(format, ...) LOG_IMPL(LogLevel_Err, format, ##__VA_ARGS__)
#define LOGD(format, ...) LOG_IMPL(LogLevel_Debug, format, ##__VA_ARGS__)

/**
 * @brief UART data handler
 * @param receiveBuffer  UART data received buffer, no need to free by user
 * @param bytesRead  The length of data received
 */
typedef void (*UART_DATA_HANDLER)(uint8_t* receiveBuffer, ssize_t bytesRead);

/**
 * @brief init system
 * 
 * @return int  0 on success, -1 otherwise
 */
int gdn_sys_init(void);

/**
 * @brief init azure client
 * 
 * @param argc 
 * @param argv 
 * @return int  0 on success, -1 otherwise
 */
int gdn_azure_init(int argc, char* argv[]);

/**
 * @brief init led
 * 
 * @return int 0 on success, -1 otherwise
 */
/**
 * @brief init led
 * 
 * @return int 0 on success, -1 otherwise
 */
int gdn_init_led(void);

/**
 * @brief led status set
 * 
 * @param type led possition
 * @param state 0 off, 1 on
 */
void gdn_set_led_status(int type, unsigned char state);

/**
 * @brief open usb to uart by params including config and data handler,This is used for the FT234XD
 * 
 * @param config Defined by UART_Config
 * @param handler Callback defined by UART_DATA_HANDLER, if set NULL, using built-in handler
 * @return int 0 on success, -1 otherwisw
 */
int gdn_open_usb_uart(UART_Config* config, UART_DATA_HANDLER handler);

/**
 * @brief gdn_close_usb_uart - close usb to uart by params including config and data handler,This is used for the FT234XD
 * 
 */
void gdn_close_usb_uart(void);

/**
 * @brief send data to usb
 * 
 * @param data data - Data buffer to send
 * @param dataLen dataLen - Length of data to send
 * @return size_t Bytes of sent data
 */
size_t gdn_send_message_to_usb(uint8_t* data, size_t dataLen);

/**
 * @brief check network ready state
 * 
 * @return int 1 Internet available, 0 Internet not available
 */
int gdn_check_network_ready(void);

/**
 * @brief get network type
 * 
 * @return e_network_type enum e_network_type
 */
e_network_type gdn_get_network_type(void);

/**
 * @brief check current wifi status
 * 
 * @return int 0 on success, -1 otherwisw
 */
int gdn_check_current_wifi_network_status(void);

/**
 * @brief set network type
 * 
 * @param networkType enum networkType
 */
void gdn_set_network_by_type(int networkType);

/**
 * @brief set network type and check
 * 
 * @param networkType enum e_network_type
 * @return int 0 on success, -1 otherwisw
 */
int gdn_change_network_type(int networkType);

/**
 * @brief Set airplane mode
 * 
 * @param onoff 0 is OFF, 1 is ON
 * @return int 0 on success, -1 otherwise
 */
int gdn_set_radio(int onoff);

/**
 * @brief Configure cellular properties: apn, uname, psw, pin, number  AND access_tech(GSM =1, LTE-M=2, LTE-NB=3)
 * 
 * @param cellular_config Attribute fields that need to be configured
 * @param access_tech  GSM =1, LTE-M=2, LTE-NB=3
 * @return int 
 */
int gdn_config_cellular(CELLULAR_CONFIG_S *cellular_config);

/**
 * @brief get esim_id:IMI,QCCID
 * 
 * @param esim_id The returned esim_id structure requires the user to allocate memory
 * @return int 0 on success, -1 otherwise
 */
int gdn_get_esim_id(ESIM_ID_S *esim_id);

/**
 * @brief get cellular status:state,signal-quality,access_tech,imei,registration-state,operator-code,operator-name
 * 
 * @param cellular_status The returned cellular_status structure requires the user to allocate memory
 * @return int 0 on success, -1 otherwise
 */
int gdn_get_cellular_status(CELLULAR_STATUS_S *cellular_status);

/**
 * @brief get gps info:utc,longitude,latitude,altitude
 * 
 * @param gps The returned gps structure requires the user to allocate memory
 * @return int 0 on success, -1 otherwise
 */
int gdn_get_gps(GPS_INFO_S *gps);

/**
 * @brief get gps SNA info:ncnt,snrs
 * 
 * @param gpsnr The returned gps structure requires the user to allocate memory
 * @return int 0 on success, -1 otherwise
 */
int gdn_get_gpsnr(GPS_SNR_INFO_S *gpsnr);

/**
 * @brief print cellular setting to debug interface
 * 
 * @return int 
 */
/**
 * @brief check if GPS signal locked 
 * 
 * @return true if locked, otherwise false
 */
bool gdn_IsGPSFixed(void);
int gdn_get_cellular_setting(void);

/**
 * @brief AT command transparent transmission to obtain corresponding response data.
 * 
 * @param atcmd AT command sent
 * @param timeout_ms Timeout waiting for reply
 * @param at_result Returned data.The user does not need to allocate memory, but the user needs to release it after use
 * @param result_len The length of the returned data
 * @return int 0 on success, -1 otherwise
 */
int gdn_handle_atcmd(char *atcmd, unsigned int timeout_ms, char **at_result, unsigned int *result_len);

/**
 * @brief One time to print detailed information such as the complete status and attributes of the cellular
 * 
 */
void gdn_print_cellular_debug_info(void);

/**
 * @brief get wwan infomation: connected,suspended,wwan_if,ip_timeout,apn,roaming,method,address,prefix,prefix,dns,mtu
 * 
 * @param wwan_info The returned wwan_info structure requires the user to allocate memory
 * @return int 0 on success, -1 otherwise
 */
int gdn_get_wwan_info(WWAN_INFO_S *wwan_info);

/**
 * @brief get nxp connect state
 * 
 * @return true connected
 * @return false not connected
 */
bool gdn_cellular_connect_state(void);

/**
 * @brief get cellular sysinfo
 * 
 * @param config The returned cellular_sys_info structure requires the user to allocate memory
 * @return int int 0 on success, -1 otherwise
 */
int gdn_get_cellular_sys_info(CELLULAR_SYS_INFO_S *config);

/**
 * @brief get cellular current config
 * 
 * @param config  The returned cellular_config structure requires the user to allocate memory
 * @return int int 0 on success, -1 otherwise
 */
int gdn_get_cellular_current_config(CELLULAR_CONFIG_S *config);

/**
 * @brief get cellular net message
 * 
 * @param config The returned net message string requires the user to allocate memory
 * @return int int 0 on success, -1 otherwise
 */
int gdn_get_cellular_netdiag(char **config);

/**
 * @brief daughter board power down
 * 
 * @return int 0 on success, -1 otherwise
 */
int gdn_daughter_board_power_up(void);

/**
 * @brief daughter board power down
 * 
 * @return int 0 on success, -1 otherwise
 */
int gdn_daughter_board_power_down(void);
void gdn_TriggerReboot(void);




