#ifndef GUARDIAN_MODBUS_H
#define GUARDIAN_MODBUS_H

#ifdef __cplusplus
extern "C"
{
#endif

/* Quantity of Coils to read  (2 bytes): 1 to 2000 (0x7D0)
 * Quantity of Coils to write (2 bytes): 1 to 1968 (0x7B0)
 */
#define MODBUS_MAX_READ_BITS              2000
#define MODBUS_MAX_WRITE_BITS             1968

 /* Quantity of Registers to read  (2 bytes): 1 to 125 (0x7D)
  * Quantity of Registers to write (2 bytes): 1 to 123 (0x7B)
  */
#define MODBUS_MAX_READ_REGISTERS          125
#define MODBUS_MAX_WRITE_REGISTERS         123

/* Consequently:
 * - RTU MODBUS ADU = 253 bytes + Server address (1 byte) + CRC (2 bytes) = 256 bytes
 * - TCP MODBUS ADU = 253 bytes + MBAP (7 bytes) = 260 bytes
 * so the maximum of both backend in 260 bytes. This size can used to allocate
 * an array of bytes to store responses and it will be compatible with the two backends
 */
#define MODBUS_MAX_ADU_LENGTH              260

/* Four arrays to store bits, input bits, registers and inputs registers
 * The pointers are stored in modbus_mapping_t structure
 */
typedef struct _modbus_mapping_t {
    int nb_bits;
    int start_bits;
    int nb_input_bits;
    int start_input_bits;
    int nb_input_registers;
    int start_input_registers;
    int nb_registers;
    int start_registers;
    uint8_t *tab_bits;
    uint8_t *tab_input_bits;
    uint16_t *tab_input_registers;
    uint16_t *tab_registers;
} modbus_mapping_t;

/* The modbus_t context is an opaque structure containing all necessary information to 
 * establish a connection with other Modbus devices according to the selected variant
 */
typedef struct _modbus modbus_t;

/* Various types of serial port enumeration values, e.g. SERIAL_RS485, SERIAL_UART.
 */
typedef enum serial_type_e
{
    SERIAL_UART  = 0,
    SERIAL_RS485 = 1,
}SERIAL_TYPE_E;

/* Master/Client Model Fuction(include fuction code 0x01-0x06).
 * Read data : gdn_mb_read_bits. gdn_mb_read_input_bits. gdn_mb_read_registers. gdn_mb_read_input_registers.
 * Write data: gdn_mb_write_bit. gdn_mb_write_register.
 */

 /* Description: Allocate and initialize a modbus_t structure to communicate with RTU mode via RS485.
  * Parameters: 1.The serial_type: Various types of serial port enumeration values, e.g. SERIAL_RS485, SERIAL_UART.
  *             2.The baud: The baud rate of the communication, e.g. 9600, 19200, 57600, 115200, etc.
  *             3.The parity: One of the following values: N for none, E for even, O for odd.
  *             4.The data_bit: The number of data bits. The allowed values are 5, 6, 7, and 8.
  *             5.The stop_bit: The stop bits which allowed values are 1 and 2.
  * Return: Return a pointer to a modbus_t structure if successful. Otherwise, that returns NULL.
  */
modbus_t* gdn_mb_serial_init(SERIAL_TYPE_E serial_type, int baud, char parity, int data_bit, int stop_bit);

/* Description: Allocate and initialize a modbus_t structure to communicate with a Modbus TCP IPv4 server.
 * Parameters: 1.The *ip: The IP address of the server to which the client wants to establish a connection.
 *             2.The port: The used TCP port.
 * Return: Return a pointer to a modbus_t structure if successful. Otherwise, that shall return NULL.
 */
modbus_t* gdn_mb_tcp_init(const char* ip, int port);

/* Description: Release an allocated modbus_t structure.
 * Parameters: 1.ctx: An allocated modbus_t structure.
 * Return: There is no return value.
 */
void gdn_mb_deinit(modbus_t* ctx);

/* Description: Reads the Boolean status of bits and sets the array elements in the destination to TRUE or FALSE (single bits).
 *              The function uses the Modbus function code 0x01 (read coil status).
 * Parameters: 1.The ctx: The gdn_mb_rs485_init. gdn_mb_uart_init. gdn_mb_tcp_init. are allocated a modbus_t structure.
 *             2.The slave_id: Define the slave ID of the remote device to talk in master mode.
 *             3.The addr: The address of the remote device.
 *             4.The nb: The number of the coil bits.
 *             5.The dest: The result of reading. Users must take care to allocate enough memory to store the results in dest (at least nb * sizeof(uint8_t)).
 * Return: Return the number of reading bits if successful. Otherwise, it shall return -1.
 */
int gdn_mb_read_bits(modbus_t* ctx, int slave_id, int addr, int nb, uint8_t* dest);

/* Description: Same as gdn_mb_read_bits but reads the remote device input table.
 *              The function uses the Modbus function code 0x02 (read input status).
 * Parameters: 1.The ctx: gdn_mb_rs485_init. gdn_mb_uart_init. gdn_mb_tcp_init. are allocated a modbus_t structure.
 *             2.The slave_id: The slave ID of the remote device to talk in master mode.
 *             3.The addr: The address of the remote device.
 *             4.The nb: The number of the coil bits.
 *             5.The dest: The result of reading. Users must take care to allocate enough memory to store the results in dest (at least nb * sizeof(uint8_t)).
 * Return: Return the number of reading input status if successful. Otherwise, it shall return -1.
 */
int gdn_mb_read_input_bits(modbus_t* ctx, int slave_id, int addr, int nb, uint8_t* dest);

/* Description: Reads the holding registers of the remote device and puts the data into an array.
 *              The function uses the Modbus function code 0x03 (read holding registers).
 * Parameters: 1.The ctx: gdn_mb_rs485_init. gdn_mb_uart_init. gdn_mb_tcp_init are allocated a modbus_t structure.
 *             2.The slave_id: Define the slave ID of the remote device to talk in master mode.
 *             3.The addr: The address of the remote device.
 *             4.The nb: The number of the holding registers.
 *             5.The dest: The result of reading. Users must take care to allocate enough memory to store the results in dest (at least nb * sizeof(uint16_t)).
 * Return: Return the number of reading registers if successful. Otherwise, that shall return -1.
 */
int gdn_mb_read_registers(modbus_t* ctx, int slave_id, int addr, int nb, uint16_t* dest);

/* Description: Reads the input registers of the remote device and put the data into an array.
 *              The function uses the Modbus function code 0x04 (read input registers).
 *              The holding registers and input registers have different historical meaning,
 *              but nowadays its more common to use holding registers only.
 * Parameters: 1.The ctx: gdn_mb_rs485_init. gdn_mb_uart_init. gdn_mb_tcp_init. are allocated a modbus_t structure.
 *             2.The slave_id: Define the slave ID of the remote device to talk in master mode.
 *             3.The addr: The address of the remote device.
 *             4.The nb: The number of the input registers.
 *             5.The dest: The result of reading. Users must take care to allocate enough memory to store the results in dest (at least nb * sizeof(uint16_t)).
 * Return: Return the number of reading input registers if successful. Otherwise, it shall return -1.
 */
int gdn_mb_read_input_registers(modbus_t* ctx, int slave_id, int addr, int nb, uint16_t* dest);

/* Description: Turns ON or OFF a single bit of the remote device.
 * The function uses the Modbus function code 0x05 (force single coil).
 * Parameters: 1.The ctx: The gdn_mb_rs485_init. gdn_mb_uart_init. gdn_mb_tcp_init. are allocated a modbus_t structure.
 *             2.The slave_id: Define the slave ID of the remote device to talk in master mode.
 *             3.The addr: The address of the remote device.
 *             4.The status: Write the status of the addr address to the remote device. The value must be set to TRUE or FALSE.
 * Return: Return 1 if successful. Otherwise, that shall return -1.
 */
int gdn_mb_write_bit(modbus_t* ctx, int slave_id, int addr, int status);

/* Description: Writes a value in one register of the remote devices.
 * Parameters: 1.The ctx: gdn_mb_rs485_init. gdn_mb_uart_init. gdn_mb_tcp_init. are allocated a modbus_t structure.
 *             2.The slave_id: Define the slave ID of the remote device to talk in master mode.
 *             3.The addr: The address of the remote device.
 *             4.The value: The value of value holding registers.
 * Return: Return 1 if successful. Otherwise, that shall return -1.
 */
int gdn_mb_write_register(modbus_t* ctx, int slave_id, int addr, const uint16_t value);

 /* Description: Write the bits of the array in the remote device.
  *              The function uses the Modbus function code 0x0F (read coil status).
  * Parameters: 1.The ctx: The gdn_mb_rs485_init. gdn_mb_uart_init. gdn_mb_tcp_init. are allocated a modbus_t structure.
  *             2.The slave_id: Define the slave ID of the remote device to talk in master mode.
  *             3.The addr: The address of the remote device.
  *             4.The nb: The number of the coil bits.
  *             5.The src: the bits of the array to write.
  * Return: Return the number of reading bits if successful. Otherwise, it shall return -1.
  */
int gdn_mb_write_bits(modbus_t *ctx, int slave_id, int addr, int nb, const uint8_t *src);

 /* Description: Write the values from the array to the registers of the remote device.
  *              The function uses the Modbus function code 0x10 (preset multiple registers).
  * Parameters: 1.The ctx: gdn_mb_rs485_init. gdn_mb_uart_init. gdn_mb_tcp_init are allocated a modbus_t structure.
  *             2.The slave_id: Define the slave ID of the remote device to talk in master mode.
  *             3.The addr: The address of the remote device.
  *             4.The nb: The number of the holding registers.
  *             5.The src: The values from the array to write.
  * Return: Return the number of reading registers if successful. Otherwise, that shall return -1.
  */
int gdn_mb_write_registers(modbus_t *ctx, int slave_id, int addr, int nb, const uint16_t *src);

/* Slave/Server Model Fuction(The server is waiting for request from clients and must answer when it is concerned by the request).
 * RTU server: gdn_mb_rs485_init. gdn_mb_uart_init.
 * TCP server: gdn_mb_tcp_accept.
 * public function: gdn_mb_mapping_new. gdn_mb_mapping_free. gdn_mb_receive. gdn_mb_reply.
 */

 /* Description: Create register memory. Allocate four arrays to store bits, input bits, registers, and inputs registers.
  *              The pointers are stored in the modbus_mapping_t structure. All values of the arrays are initialized to zero.
  *              If it isn't necessary to allocate an array for a specific type of data, users can pass the zero value in argument,
  *              the associated pointer will be NULL.
  * Parameters: 1.The nb_bits: The number of coils.
  *             2.The nb_input_bits: The number of input channels.
  *             3.The nb_registers: The number of holding registers.
  *             4.The nb_input_registers: The number of input registers.
  * Return: Return the new allocated structure if successful. Otherwise, that shall return NULL.
  */
modbus_mapping_t* gdn_mb_mapping_new(int nb_bits, int nb_input_bits, int nb_registers, int nb_input_registers);

/* Description: Free the four arrays of the mb_mapping_t structure.
 * Parameters: The mb_mapping: An allocated modbus_mapping_t structure.
 * Return: There is no return value.
 */
void gdn_mb_mapping_free(modbus_mapping_t* mb_mapping);

/* Description: listen to maximum nb_connection incoming connections on the specified IP address. Block waiting for client connection. Add port for permissions in app_manifest.json.
 * Parameters: 1.The ip: The IP address of the server. A NULL value can be used to listen to any addresses.
 *             2.The port: The TCP server allowed the port to use.
 *             3.The nb_connection: The maximum incoming connections on the specified IP address.
 *             4.The socket: Create a new socket, but need to close the socket after use.
 * Return: Return a pointer to a modbus_t structure if successful. Otherwise, that shall return NULL.
 */
modbus_t* gdn_mb_tcp_accept(const char* ip, int port, int nb_connection, int* socket);

/* Description: This function is used by the Modbus slave/server to receive and analyze indication requests sent by the masters/clients.
 * Parameters: 1.The ctx: The gdn_mb_tcp_accept. gdn_mb_rs485_init. gdn_mb_uart_init. are allocated a modbus_t structure.
 *             2.The self_id: The internal slave ID in slave mode.
 *             3.The req: An indication request from the context ctx.
 * Return: The indication request in req and return the request length if successful.
 *         The returned request length can be zero if the indication request is ignored (e.g. query for another slave in RTU mode).
 *         Otherwise, it shall return -1.
 */
int gdn_mb_receive(modbus_t* ctx, int self_id, uint8_t* req);

/* Description: send a response to the received request.
 * Parameters: 1.The ctx: The gdn_mb_receive. The information of the Modbus context ctx
 *             2.The req: The gdn_mb_receive. Fill in a response to the received request.
 *             3.The req_length: The length of the req argument
 *             4.The mb_mapping: An allocated modbus_mapping_t structure. If the request indicates to read or write a value, the operation will be done in the Modbus mapping according to the type of the manipulated data.
 * Return: Return the length of the response sent if successful. Otherwise, it shall return -1
 */
int gdn_mb_reply(modbus_t* ctx, const uint8_t* req, int req_length, modbus_mapping_t* mb_mapping);

#ifdef __cplusplus
}
#endif

#endif
